/*
 * File: Shape.c
 */
#include <stdio.h>
#include <stdlib.h>
#include "Shape.h"
#include "ShapeProtected.h" // internals, the "pimpl" in C++ jargon :)

// Interface function implementations: have access to the internals

/*
 * Virtual functions
 * -----------------
 * They must comply with signatures of functions in struct ShapeInterfaceT.
 * The "instance" argument in all of these must be NULL for consistency, because
 * the fact that we address them means that the "child" object has not been
 * initialized yet. To make sure, one can implement an extra check (which we didn't).
 */

#ifdef CHECK_INTERNAL
#define CHECK_NULL_INSTANCE						\
	if(instance) {							\
		fprintf(stderr, "error: internal inconsistency!\n");	\
		exit(1); }
#else
#define CHECK_NULL_INSTANCE
#endif

// can probably write a macro for all of these

static float VirtualShapeArea(void *instance) {
	CHECK_NULL_INSTANCE
	fprintf(stderr,"error: call to virtual function ShapeArea\n");
	#ifdef EXIT_ON_VIRTUAL_CALL
		exit(1);
	#else
		return 0.;
	#endif
}

static float VirtualShapePerimeter(void *instance) {
	CHECK_NULL_INSTANCE
	fprintf(stderr,"error: call to virtual function ShapePerimeter\n");
	#ifdef EXIT_ON_VIRTUAL_CALL
		exit(1);
	#else
		return 0.;
	#endif
}

static void VirtualShapeDraw(void *instance, int color) {
	CHECK_NULL_INSTANCE
	fprintf(stderr,"error: call to virtual function ShapeDraw\n");
	#ifdef EXIT_ON_VIRTUAL_CALL
		exit(1);
	#endif
}

static void VirtualShapeDestruct(void *instance) {
	CHECK_NULL_INSTANCE
	fprintf(stderr,"error: call to virtual function ShapeDestruct\n");
	#ifdef EXIT_ON_VIRTUAL_CALL
		exit(1);
	#endif
}

static struct ShapeClassMembers ShapeVFT = {
	.class_name = "Shape", // or could be just NULL
	.obj_counter = 0, // we count the virtual shapes separately
	.area = VirtualShapeArea, 
	.perimeter = VirtualShapePerimeter,
	.draw = VirtualShapeDraw,
	.destruct = VirtualShapeDestruct
};

// Constructor
static void ShapeReset(Shape shape) {
	// Auxiliary function
	shape->instance = NULL;
	shape->vft = &ShapeVFT;
}

// "New" calls like this should match "Delete" + "Replace" calls
Shape ShapeNew(float x, float y) {
	Shape shape = malloc(sizeof(struct ShapeInterfaceT));
	ShapeReset(shape);
	shape->vft->obj_counter++; // counting virtual shapes only
	shape->x = x; shape->y = y;
	return shape;
}

// Destructor (the derived class destructors are hidden from the user)
// "New" calls should match "Delete" calls like this + "Replace" calls
void ShapeDelete(Shape *shapePtr) {
	// We don't check if shapePtr is NULL, we should not use addresses to anything but Shape
	Shape shape = *shapePtr;
	if (!shape) {
		fprintf(stderr,"error: double delete of the shape interface\n");
		#ifdef EXIT_ON_DOUBLE_DELETE
			exit(1);
		#else
			return;
		#endif
	}
	void *inst = shape->instance;
	if (inst) {
		shape->vft->destruct(inst);  // destruct the derived part
		// Note that object counting CANNOT occur here because Shape does not have
		// access to DerivedCreate which increases counter, so that the opposite
		// operation is impossible.
		ShapeReset(shape); // important, restore virtual shape class members
	} // if inst==NULL, then it means that some other interface already destructed it
	shape->vft->obj_counter--; // counting virtual shapes only
	free(shape);
	*shapePtr = NULL; // this way we can tell it does not point to anything anymore
}

// Equivalent of the copy-assign operator in C++
void ShapeReplace(Shape *old_shapePtr, Shape new_shape) {
	ShapeDelete(old_shapePtr);
	*old_shapePtr = new_shape;
}

// To check if it is a derived class (DEBUGGING ONLY!)
bool ShapeIsVirtual(Shape shape) { return (shape->instance == NULL); }

/*
 * The interface
 */

// Methods
float ShapeArea(Shape shape) { return shape->vft->area(shape->instance); }
float ShapePerimeter(Shape shape) { return shape->vft->perimeter(shape->instance); }
void ShapeDraw(Shape shape,int color) { shape->vft->draw(shape->instance, color); }

// Data member access (getters and setters), making them available to the user
float ShapeGetX(Shape shape) { return shape->x; }
float ShapeGetY(Shape shape) { return shape->y; }
void ShapeSetX(Shape shape, float new_x) { shape->x = new_x; }
void ShapeSetY(Shape shape, float new_y) { shape->y = new_y; }

// Class member access
char *ShapeGetClassName(Shape shape) { return shape->vft->class_name; }
int ShapeCounterOfThisType(Shape shape) { return shape->vft->obj_counter; }
int ShapeCounterTotal(void) { return ShapeVFT.obj_counter; }

// More methods
void ShapeTranslation(Shape shape, float dx, float dy) {
	shape->x += dx; shape->y += dy;
	printf("Moving shape by (dx=%f, dy=%f) to the new position at (%f, %f)\n", dx, dy, shape->x, shape->y);
}

