/*
 * File: Smiley.c
 * --------------
 * Like others, inherits from Shape.
 * However, does not inherit from Circle.
 * Instead, has a circle as an element for "face".
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h> /* bool, true and false */
#include "Shape.h"
#include "ShapeProtected.h"
#include "Circle.h"
#include "Rectangle.h"
#include "Smiley.h"

#define NFACEELEM 4
struct SmileyCDT {
	Shape shape;
	Shape elements[NFACEELEM]; // Face, eyes and mouth
	bool isForeground[NFACEELEM]; // true if foreground, false if background
	Circle face; // composition over inheritance!
	Circle eyes[2];
	Rectangle mouth;
};

Shape SmileyShape(Smiley smiley) { return smiley->shape; }

// We access these functions only through the "shape" interface
static float SmileyArea(void *instance) {
	Smiley smiley = (Smiley) instance;
	float area = 0;
	int i;
	for (i = 0; i < NFACEELEM; i++)
		// areas of eyes and mouth are subtracted
		area += (smiley->isForeground[i] ? 1. : -1.)*ShapeArea(smiley->elements[i]);
	return area;
}

static float SmileyPerimeter(void *instance) {
	Smiley smiley = (Smiley) instance;
	float perimeter = 0;
	int i;
	for (i = 0; i < NFACEELEM; i++)
		perimeter += ShapePerimeter(smiley->elements[i]);
	return perimeter;
}

static void SmileyDraw(void *instance, int color) {
	Smiley smiley = (Smiley) instance;
	int i;
	printf("I am a Smiley of radius %f at (%f, %f) drawn with color %d!\n",
		CircleGetRadius(smiley->face), smiley->shape->x, smiley->shape->y, color);
	for (i = 0; i < NFACEELEM; i++) {
		printf(" - Drawing smiley element %d: ",i);
		ShapeDraw(smiley->elements[i], smiley->isForeground[i]*color);
	}
}

static void SmileyDestruct(void *instance) {
	Smiley smiley = (Smiley) instance;
	Shape shape = smiley->shape;
	int i;
	#ifndef QUIET_DTOR
		printf("Destructing smiley at (%f,%f)\n", shape->x, shape->y);
	#endif
	shape->vft->obj_counter--;
	for (i = 0; i < NFACEELEM; i++) {
		#ifndef QUIET_DTOR
			printf(" - Destructing smiley element %d: ",i);
		#endif
		ShapeDelete(&(smiley->elements[i]));
	}
	free(smiley);
	// Keep "shape", this method destructs only the "derived" stuff
}

static struct ShapeClassMembers SmileyVFT = {
	.class_name = "Smiley",
	.obj_counter = 0,
	.area = SmileyArea, 
	.perimeter = SmileyPerimeter,
	.draw = SmileyDraw,
	.destruct = SmileyDestruct
};

Smiley SmileyCreate(Shape shape, float radius) {
	int i;
	double x, y;
	Smiley smiley = malloc(sizeof(struct SmileyCDT));
	// Attach the base (this must be in every derived class)
	if (shape) smiley->shape = shape;
	else { fprintf(stderr,"error: base not initalized in SmileyCreate\n"); exit(1); }
	shape->instance = smiley;
	// Overload the inherited members (virtual functions)
	shape->vft = &SmileyVFT;
	// Done with preliminaries! Now, the proper initialization
	// 1. Do whatever is needed with derived class members
	shape->vft->obj_counter++;
	// 2. Initialize the derived object members
	x = shape->x; y = shape->y;
	#ifndef QUIET_CTOR
		printf("Creating a Smiley at (%f, %f) of radius=%f\n", x, y, radius);
	#endif
	smiley->face = CircleCreate(ShapeNew(x, y), radius);
	smiley->eyes[0] = CircleCreate(ShapeNew(x - radius/3.,y), radius/10.);
	smiley->eyes[1] = CircleCreate(ShapeNew(x + radius/3.,y), radius/10.);
	smiley->mouth = RectangleCreate(ShapeNew(x, y-radius/2.), radius/4., radius/20.);
	// Associate the shapes with elements
	smiley->elements[0] = CircleShape(smiley->face);
	smiley->elements[1] = CircleShape(smiley->eyes[0]);
	smiley->elements[2] = CircleShape(smiley->eyes[1]);
	smiley->elements[3] = RectangleShape(smiley->mouth);
	smiley->isForeground[0] = true;
	for (i = 1; i < NFACEELEM; i++) smiley->isForeground[i] = false;
	return smiley;
}

// Smiley methods

int SmileyCounter(void) { return SmileyVFT.obj_counter; }
float SmileyGetRadius(Smiley smiley) { return CircleGetRadius(smiley->face); }
void SmileySetRadius(Smiley smiley, float new_radius) {
	fprintf(stderr,"error: resizing smileys is not implemented\n");
	//circle->radius = new_radius;
}

