/*
 * File: main.c
 * ------------
 * Implementation of "inheritance" pattern (or, equivalently, an "interface") in C.
 * We demonstrate that it is possible without introducing a mess with "switch" statements.
 * We keep the OOP features such as encapsulation.
 * 
 * To avoid memory leaks, the all "*New" operations must be matched with "*Delete" or "*Replace"
 * operations. This does not apply to "*Create" operations because the corresponding "*Destruct"
 * is called automatically.
 *
 * Project compilation:
 *     make
 * or just simply
 *     gcc -Wall main.c Shape.c Circle.c Rectangle.c Smiley.c -o inheritance.out
 * Run it:
 *     ./inheritance.out
 *
 * The virtual function call is an error. However, by default we choose to continue execution.
 * To exit on such an error, do:
 *     make OPT="-DEXIT_ON_VIRTUAL_CALL"
 *
 * For additional testing do
 *     make OPT="-DCREATE_OBJECT_FIRST"
 *
 * For extra debugging:
 *     make OPT="-DCHECK_INTERNAL"
 *
 * To suppress info output about creator and destructor calls:
 *     make OPT="-DQUIET_CTOR -DQUIET_DTOR"
 *
 * The options may be combined, e.g.:
 *     make OPT="-DEXIT_ON_VIRTUAL_CALL -DCREATE_OBJECT_FIRST"
 *
 * However, make sure to call "make clean" before recompiling with changed options!!!
 */
#include <stdio.h>
#include <stdlib.h>

#ifdef MEMORY_CHECK
#include <mcheck.h>
void no_op(enum mcheck_status status) {}
#endif

#include "Shape.h"
#include "Circle.h"
#include "Rectangle.h"
#include "Smiley.h"

#define N 5

int main(void)
{
	#ifdef MEMORY_CHECK
		mcheck(&no_op);
	#endif
	int i;
	float radius, width, height, x, y; // readability variables
	Shape shapes[N];
	// Prepare for possible mess
	for (i = 0; i < N; i++) shapes[i] = NULL;

	// 1. CREATION
	printf("\n*** Creating shapes ***\n");
	// Initialize the shapes first
	for (i = 0; i < N; i++) {
		shapes[i] = ShapeNew(x=10.*i+1.2, y=10.*i+2.4);
	}
	// Concrete objects
	#ifdef MEMORY_CHECK
		// Saving objects may be actually dangerous if we delete the underlying shape!
		Circle savedCircle = CircleCreate(shapes[0], radius=2.);
		Shape savedShape = CircleShape(savedCircle);
	#else
		CircleCreate(shapes[0], radius=2.); // Do not save by default!
	#endif
	// Usually we don't save the objects, because we only use their ability to interface
	RectangleCreate(shapes[1], width=3., height=4.);
	SmileyCreate(shapes[2], radius=5.);
	// This may be done the other way around, the object may be created first and then its base be added to the array:
	#ifdef CREATE_OBJECT_FIRST
		printf("*** EXTRA: Creating the object first, then getting access to base for Circle\n");
		Circle circle = CircleCreate(ShapeNew(x=20., y=30.), radius=6.);
		// Arbitrarily shift the circle using the "Shape" interface
		float dx, dy; // readability variables
		ShapeTraslation(CircleShape(circle), dx = 1., dy = 2.);
		// Store the new shape in the array, while replacing the old one.
		// These two operations are equivalent to the copy-assignment operator in C++:
		//   ShapeDelete(&shapes[0]); // prevent memory leaks; also destructs the derived object. Matches "ShapeNew" above
		//   shapes[0] = CircleShape(circle); // put the new one into the array
		ShapeReplace(&shapes[0], CircleShape(circle)); // combine into one operation
		// Note that these operations invalidate savedCircle!
	#endif
	// No concrete object are created for shapes[2], shapes[3], to test error recovery

	// 2. USER USE
	printf("\n*** Drawing the shapes ***\n");
	for (i = 0; i < N; i++) {
		Shape shape = shapes[i];
		if (!shape) continue;
		printf("\n");
		if (ShapeIsVirtual(shape)) printf("NOTE: Virtual shape but still trying to continue\n");
		int color = i+1; // just to test
		// It is possible to use a simplified interface for non-virtual and data members, e.g., shape->x, shape->y
		// For this, the internals of the interface in "ShapeProtected.h" must be made available here.
		// But we use getters/setters so that the internals of Shape are hidden from user completely
		printf("Drawing shape #%d (%s) at (%f,%f) of color %d\n",
			i,ShapeGetClassName(shape),ShapeGetX(shape),ShapeGetY(shape),color);
		// The virtual (overloaded) methods have an extra layer of interface
		printf("This shape: area=%f, perimeter=%f\n", ShapeArea(shape), ShapePerimeter(shape));
		ShapeDraw(shape,color);
		printf("There are %d objects of this shape (%s)\n",ShapeCounterOfThisType(shape),ShapeGetClassName(shape));
	}

	// 3. DESTRUCTION
	printf("\nThe class counters are:\n\tShape = %d (total)\n"
		"\tCircle = %d\n\tRectangle = %d\n\tSmiley = %d\n",
		ShapeCounterTotal(),CircleCounter(),RectangleCounter(),SmileyCounter());
	printf("\n*** Destructing the shapes ***\n");
	#ifdef MEMORY_CHECK
		printf("The mcheck_status values:\n\tMCHECK_DISABLED = %d\n\tMCHECK_OK       = %d\n"
			"\tMCHECK_FREE     = %d\n\tMCHECK_HEAD     = %d\n\tMCHECK_TAIL     = %d\n",
			MCHECK_DISABLED, MCHECK_OK, MCHECK_FREE, MCHECK_HEAD, MCHECK_TAIL);
		// The following line shows that savedCircle is invalid if CREATE_OBJECT_FIRST
		// block was executed (it was destructed and memory freed!)
		printf("The saved circle: %p, status=%d\n",savedCircle,mprobe(savedCircle)); fflush(stdout);
		printf("The saved circle shape: %p, status=%d\n",savedShape,mprobe(savedShape));
	#endif
	#if 0
		// Destruct a circle explicitly -- demo of what not to do!
		if (savedCircle) {
			// this operation puts us in a world of pain because the memory may have been already freed!
			CircleDelete(&savedCircle);
		} else {
			printf("Circle already deleted\n"); // when it was replace in the block CREATE_OBJECT_FIRST
		}
	#endif
	#ifdef MEMORY_CHECK
		printf("The objects to be destructed are: [");
		for (i = 0; i < N; i++) {
			Shape shape = shapes[i];
			printf("\n\t%p:virt=%d,status=%d; ", shape, shape?ShapeIsVirtual(shape):-1,mprobe((void*)shape));
		}
		printf("]\n");
		fflush(stdout);
	#endif
	for (i = 0; i < N; i++) { ShapeDelete(&(shapes[i])); }
	// The shapes are now NULL, to prevent future errors
	// Check if we are done:
	printf("The destructed objects are: [");
	for (i = 0; i < N; i++) printf("%p, ", shapes[i]);
	printf("]\n");
	printf("\nThe class counters are:\n\tShape = %d (total)\n"
		"\tCircle = %d\n\tRectangle = %d\n\tSmiley = %d\n",
		ShapeCounterTotal(),CircleCounter(),RectangleCounter(),SmileyCounter());
	return 0;
}

